/*
 * yasapi_ring_conv.c
 *
 * Yet Another (WA)SAPI (YASAPI) New Technology (NT) Output (out_yasapi-nt)
 * Copyright (C) 2015-2018  Peter Belkner (pbelkner@snafu.de)
 *
 * out_yasapi-nt is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * out_yasapi-nt is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with out_yasapi-nt.  If not, see <https://www.gnu.org/licenses/>.
 */
#include <yasapi.h>
#include <stdint.h>
#include <math.h>

///////////////////////////////////////////////////////////////////////////////
static DWORD RingConvertConvertSizeSource(Ring *pRing, DWORD dwSize)
{
  Engine *pEngine=pRing->pEngine;
  WAVEFORMATEX *pTarget=&pEngine->wfx.target.Format;
  WAVEFORMATEX *pSource=&pEngine->wfx.source.Format;

  dwSize=MulDiv(dwSize,pSource->wBitsPerSample,pTarget->wBitsPerSample);

  if (!pEngine->options.common.eMono2Stereo)
    goto exit;

  if (1u!=pSource->nChannels||2u!=pTarget->nChannels)
    goto exit;

  dwSize>>=1ul;
exit:
  return dwSize;
}

static DWORD RingConvertSizeSource(Ring *pRing)
{
  return RingConvertConvertSizeSource(pRing,pRing->_mp-pRing->_buf);
}

static DWORD RingConvertWrittenSource(Ring *pRing)
{
  return RingConvertConvertSizeSource(pRing,pRing->_dwWritten);
}

static DWORD RingConvertAvailableSource(Ring *pRing)
{
  return RingConvertConvertSizeSource(pRing,pRing->_dwAvailable);
}

////////////////
static DWORD RingConvertConvertSizeTarget(Ring *pRing, DWORD dwSize)
{
  Engine *pEngine=pRing->pEngine;
  WAVEFORMATEX *pTarget=&pEngine->wfx.target.Format;
  WAVEFORMATEX *pSource=&pEngine->wfx.source.Format;

  dwSize=MulDiv(dwSize,pTarget->wBitsPerSample,pSource->wBitsPerSample);

  if (!pEngine->options.common.eMono2Stereo)
    goto exit;

  if (1u!=pSource->nChannels||2u!=pTarget->nChannels)
    goto exit;

  dwSize<<=1ul;
exit:
  return dwSize;
}

static DWORD RingConvertSizeTarget(Ring *pRing)
{
  return pRing->_mp-pRing->_buf;
}

static DWORD RingConvertWrittenTarget(Ring *pRing)
{
  return pRing->_dwWritten;
}

#if defined (YASAPI_TARGET_AVAILABLE) // [
static DWORD RingConvertAvailableTarget(Ring *pRing)
{
  return pRing->_dwAvailable;
}
#endif // ]

////////////////
static double xConvertSqrt2Inverse=0.0;

static BYTE *RingConvertCopyToRing(Ring *pRing, BYTE *pbDst,
    const BYTE *pbSrc, DWORD dwSize)
{
  Engine *pEngine=pRing->pEngine;
  WAVEFORMATEX *pTarget=&pEngine->wfx.target.Format;
  WORD wBytesPerSampleTarget=pTarget->wBitsPerSample>>3;
  WAVEFORMATEX *pSource=&pEngine->wfx.source.Format;
  WORD wBytesPerSampleSource=pSource->wBitsPerSample>>3;
  BYTE *pbDstMin=pbDst;
  BYTE *pbDstMax=pbDst+dwSize;
  WORD wChannel;

  int32_t i;
  BYTE *iwp=(BYTE *)&i+((sizeof i)-wBytesPerSampleSource);
  const BYTE *irp=(BYTE *)&i+((sizeof i)-wBytesPerSampleTarget);

  if (pTarget->nChannels==2&&pSource->nChannels==1) {
    switch (pEngine->options.common.eMono2Stereo) {
    case eMono2StereoHalf:
      while (pbDst<pbDstMax) {
        i=0;
        memcpy(iwp,pbSrc,wBytesPerSampleSource);
        pbSrc+=wBytesPerSampleSource;

        i>>=1;

        memcpy(pbDst,irp,wBytesPerSampleTarget);
        pbDst+=wBytesPerSampleTarget;
        memcpy(pbDst,irp,wBytesPerSampleTarget);
        pbDst+=wBytesPerSampleTarget;
      }

      break;
    case eMono2StereoSqrt2Inverse:
      if (0.0==xConvertSqrt2Inverse)
        xConvertSqrt2Inverse=1.0/sqrt(2.0);

      while (pbDst<pbDstMax) {
        i=0;
        memcpy(iwp,pbSrc,wBytesPerSampleSource);
        pbSrc+=wBytesPerSampleSource;

        i=xConvertSqrt2Inverse*i+0.5;

        memcpy(pbDst,irp,wBytesPerSampleTarget);
        pbDst+=wBytesPerSampleTarget;
        memcpy(pbDst,irp,wBytesPerSampleTarget);
        pbDst+=wBytesPerSampleTarget;
      }

      break;
    default:
      if (wBytesPerSampleTarget==wBytesPerSampleSource) {
        while (pbDst<pbDstMax) {
          memcpy(pbDst,pbSrc,wBytesPerSampleTarget);
          pbDst+=wBytesPerSampleTarget;
          memcpy(pbDst,pbSrc,wBytesPerSampleTarget);
          pbDst+=wBytesPerSampleTarget;

          pbSrc+=wBytesPerSampleTarget;
        }
      }
      else {
        while (pbDst<pbDstMax) {
          i=0;
          memcpy(iwp,pbSrc,wBytesPerSampleSource);
          pbSrc+=wBytesPerSampleSource;

          memcpy(pbDst,irp,wBytesPerSampleTarget);
          pbDst+=wBytesPerSampleTarget;
          memcpy(pbDst,irp,wBytesPerSampleTarget);
          pbDst+=wBytesPerSampleTarget;
        }
      }

      break;
    }
  }
  else if (pTarget->nChannels!=pSource->nChannels) {
    DMESSAGE("Multi channel mixing not supported");
    goto exit;
  }
  else {
    while (pbDst<pbDstMax) {
      for (wChannel=0u;wChannel<pTarget->nChannels;++wChannel) {
        i=0;

        memcpy(iwp,pbSrc,wBytesPerSampleSource);
        pbSrc+=wBytesPerSampleSource;

        memcpy(pbDst,irp,wBytesPerSampleTarget);
        pbDst+=wBytesPerSampleTarget;
      }
    }
  }
exit:
  return pbDstMin;
}

static BYTE *RingConvertCopyFromRing(Ring *pRing,
    BYTE *pbDst, const BYTE *pbSrc, DWORD dwSize)
{
  return memcpy(pbDst,pbSrc,dwSize);
}

////////////////
IRingVtbl gcIRingConvertVtbl={
#if defined (YASAPI_SOURCE_CONVERT_SIZE) // [
  .source.ConvertSize=RingConvertConvertSizeSource,
#endif // ]
  .source.Size=RingConvertSizeSource,
  .source.Written=RingConvertWrittenSource,
  .source.Available=RingConvertAvailableSource,
  ////////////
  .target.ConvertSize=RingConvertConvertSizeTarget,
  .target.Size=RingConvertSizeTarget,
  .target.Written=RingConvertWrittenTarget,
#if defined (YASAPI_TARGET_AVAILABLE) // [
  .target.Available=RingConvertAvailableTarget,
#endif // ]
  ////////////
  .CopyToRing=RingConvertCopyToRing,
  .CopyFromRing=RingConvertCopyFromRing,
};
